# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

# shortcut imports
from typegraph import types as t  # noqa
from typegraph.graph.typegraph import TypeGraph  # noqa

version = "0.0.3-dev.5"
